#include "log.h"
#include <ctype.h>
#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include "selectiontree.h"
#include <set>

#ifndef min
template <class T> static inline T min(T x,T y) { return (x>y)?y:x; }
#endif

char *line;
int max_line_len;

struct id_and_value
{
	int id;
	double value;
};

int compare_id_and_value(const void *a, const void *b)
{
	struct id_and_value *ia = (struct id_and_value *)a;
	struct id_and_value *ib = (struct id_and_value *)b;
	if(ia->value > ib->value)
		return 1;
	if(ia->value < ib->value)
		return -1;
	return 0;
}

void _group_queries(const int *query_id, int l, int *nr_query_ret, int **start_ret, int **count_ret, int *perm)
{
	int max_nr_query = 16;
	int nr_query = 0;
	int *query = new int[max_nr_query];
	int *count = new int[max_nr_query];
	int *data_query = new int[l];
	int i;

	for(i=0;i<l;i++)
	{
		int this_query = (int)query_id[i];
		int j;
		for(j=0;j<nr_query;j++)
		{
			if(this_query == query[j])
			{
				++count[j];
				break;
			}
		}
		data_query[i] = j;
		if(j == nr_query)
		{
			if(nr_query == max_nr_query)
			{
				max_nr_query *= 2;
				query = (int *)realloc(query,max_nr_query * sizeof(int));
				count = (int *)realloc(count,max_nr_query * sizeof(int));
			}
			query[nr_query] = this_query;
			count[nr_query] = 1;
			++nr_query;
		}
	}

	int *start = new int[nr_query];
	start[0] = 0;
	for(i=1;i<nr_query;i++)
		start[i] = start[i-1] + count[i-1];
	for(i=0;i<l;i++)
	{
		perm[start[data_query[i]]] = i;
		++start[data_query[i]];
	}
	start[0] = 0;
	for(i=1;i<nr_query;i++)
		start[i] = start[i-1] + count[i-1];

	*nr_query_ret = nr_query;
	*start_ret = start;
	*count_ret = count;
	free(query);
	free(data_query);
}
static char* readline(FILE *input)
{
	int len;

	if(fgets(line,max_line_len,input) == NULL)
		return NULL;

	while(strrchr(line,'\n') == NULL)
	{
		max_line_len *= 2;
		line = (char *) realloc(line,max_line_len);
		len = (int) strlen(line);
		if(fgets(line+len,max_line_len-len,input) == NULL)
			break;
	}
	return line;
}

void eval(double *label, double *target, int *query, int l, double *apa, double *mean_ndcg, double *map)
{
	int q,i,j,k;
	int nr_query;
	int *start = NULL;
	int *count = NULL;
	int *perm = new int[l];
	id_and_value *pi;
	int true_query;
	long long totalnc = 0, totalnd = 0;
	long long nc = 0;
	long long nd = 0;
	double tmp;
	double accuracy = 0;
	int *l_plus;
	int *int_y;
	int same_y = 0;
	double *idcg;
	double *dcg;
	double idcg_yahoo,dcg_yahoo,size_yahoo;
	double ndcg_yahoo = 0;
	double meanndcg = 0;
	double tmp_ndcg;
	long long pairs = 0;
	selectiontree *T;
	_group_queries(query, l, &nr_query, &start, &count, perm);
	true_query = nr_query;
#ifdef DEBUG
	printf("Total Queries = %d\n", nr_query);
#endif
	for (q=0;q<nr_query;q++)
	{
		//pairwise accuracy
		nc = 0;
		nd = 0;
		l_plus = new int[count[q]];
		int_y = new int[count[q]];
		pi = new id_and_value[count[q]];
		for (i=0;i<count[q];i++)
		{
			pi[i].id = i;
			pi[i].value = label[perm[i + start[q]]];
		}
		qsort(pi, count[q], sizeof(id_and_value), compare_id_and_value);
		int_y[pi[0].id] = 1;
		same_y = 0;
		k = 1;
		for(i=1;i<count[q];i++)
		{
			if (pi[i-1].value < pi[i].value)
			{
				same_y = 0;
				k++;
			}
			else
				same_y++;
			int_y[pi[i].id] = k;
			nc += (i - same_y);
		}
		pairs += nc;
		for (i=0;i<count[q];i++)
		{
			pi[i].id = i;
			pi[i].value = target[perm[i + start[q]]];
		}
		qsort(pi, count[q], sizeof(id_and_value), compare_id_and_value);
		//total pairs
		T = new selectiontree(k);
		j = 0;
		for (i=0;i<count[q];i++)
		{
			while (j<count[q] && ( pi[j].value < pi[i].value))
			{
				T->insert_node(int_y[pi[j].id], tmp);
				j++;
			}
			T->count_larger(int_y[pi[i].id], &l_plus[pi[i].id], &tmp);
		}
		delete T;

		for (i=0;i<count[q];i++)
			nd += l_plus[i];
		nc -= nd;
		if (nc != 0 || nd != 0)
			accuracy += double(nc)/double(nc+nd);
		else
			true_query--;
		totalnc += nc;
		totalnd += nd;
		delete[] l_plus;
		delete[] int_y;
		delete[] pi;
	}
#ifdef DEBUG
	printf("Total pairs = %lld\n",pairs);
	printf("Pairwise accuracy = %g%% (ranking)\n",(double)totalnc/(double)(totalnc+totalnd)*100);
#endif
	*apa = (double)totalnc/(double)(totalnc+totalnd)*100;

	std::set<double> relevances;
	for (q = 0; q < nr_query; q++)
		for (i = 0; i < count[q]; i++)
			relevances.insert(label[i + start[q]]);
	std::vector<double> ttmp(relevances.begin(), relevances.end());
#ifdef DEBUG
	for (std::vector<double>::const_iterator it = ttmp.begin(); it != ttmp.end(); it++)
		printf("%lf ", *it);
	puts("");
#endif
	double threshold = ttmp[ttmp.size() / 2];
#ifdef DEBUG
	printf("Threshold = %g\n", threshold);
#endif
	double MAP = 0.0;
	for (q = 0; q < nr_query; q++)
	{
		pi = new id_and_value[count[q]];
		for (i = 0; i < count[q]; i++)
		{
			pi[i].id = perm[i + start[q]];
			pi[i].value = -target[pi[i].id];
		}
		qsort(pi, count[q], sizeof(id_and_value), compare_id_and_value);
		double AP = 0.0;
		int hit = 0;
		int total = 0;
		for (i = 0; i < count[q]; i++)
		{
			if (label[pi[i].id] >= threshold)
			{
				total++;
				hit++;
				AP += 1.0 * hit / (i + 1);
			}
		}
		if (total != 0)
		{
			AP /= total;
			MAP += AP;
		}
		else
			MAP += 1;
		delete [] pi;
	}

	MAP /= nr_query;
	*map = MAP;
#ifdef DEBUG
	printf("Mean Average Precision = %g \n", MAP);
#endif

	for (q=0;q<nr_query;q++)
	{
		//mean ndcg
		idcg = new double[count[q]];
		dcg = new double[count[q]];
		size_yahoo = min(count[q],10);
		tmp_ndcg = 0;
		pi = new id_and_value[count[q]];
		for (i=0;i<count[q];i++)
		{
			pi[i].id = perm[i + start[q]];
			pi[i].value = label[perm[i + start[q]]];
		}
		qsort(pi, count[q], sizeof(id_and_value), compare_id_and_value);
		idcg_yahoo = 0;
		dcg_yahoo = 0;
		idcg[0] = pow(2.0,pi[count[q]-1].value) - 1;
		for (i=1;i<count[q];i++)
			idcg[i] = idcg[i-1] + (pow(2.0,pi[count[q]-1 - i].value) - 1) * log(2.0) / log(i+1.0);
		for (i=0;i<size_yahoo;i++)
			idcg_yahoo += (pow(2.0,pi[count[q]-1 - i].value) - 1) * log(2.0) / log(i+2.0);
		for (i=0;i<count[q];i++)
		{
			pi[i].id = perm[i + start[q]];
			pi[i].value = target[perm[i + start[q]]];
		}
		qsort(pi, count[q], sizeof(id_and_value), compare_id_and_value);
		dcg[0] = pow(2.0, label[pi[count[q] - 1].id]) - 1;
		for (i=1;i<count[q];i++)
			dcg[i] = dcg[i-1] + (pow(2.0, label[pi[count[q] - 1 - i].id]) - 1) * log(2.0) / log(i + 1.0);
		for (i=0;i<size_yahoo;i++)
			dcg_yahoo += (pow(2.0,label[pi[count[q]-1 - i].id]) - 1) * log(2.0) / log(i+2.0);
		if (idcg[0]>0)
			for (i=0;i<count[q];i++)
				tmp_ndcg += dcg[i]/idcg[i];
		else
			tmp_ndcg = 0;
		meanndcg += tmp_ndcg/count[q];

		if (idcg_yahoo > 0)
			ndcg_yahoo += dcg_yahoo / idcg_yahoo;
		else
			ndcg_yahoo += 1;
		delete[] pi;
		delete[] idcg;
		delete[] dcg;
	}
	meanndcg /= nr_query;
	ndcg_yahoo /= nr_query;
	*mean_ndcg = meanndcg;
#ifdef DEBUG
	printf("MeanNDCG (LETOR) = %g (ranking)\n", meanndcg);
	printf("NDCG (YAHOO) = %g (ranking)\n", ndcg_yahoo);
#endif
	free(start);
	free(count);
	free(perm);
}

void _exit_input_error(int line)
{
	assert(false);
}

Logger::~Logger()
{
	delete [] m_decision_value;
	for (int i = 0; i < m_test_data_size; i++)
		delete [] m_test_data[i];

	delete [] m_test_data;
	delete [] m_qid;
	delete [] m_target_label;
}


Logger::Logger(const char* test_file_path, const svm_problem *problem, const svm_parameter &parameter)
{

	//record parameter and training vectors
	m_problem = problem;
	m_gamma = parameter.gamma;

	//load test file
	FILE *input;

	input = fopen(test_file_path, "r");
	assert(input != NULL);

	int total = 0;
	max_line_len = 1024;
	int max_nr_attr = 1;
	line = (char *)malloc(max_line_len*sizeof(char));
	while(readline(input) != NULL)
		total++;
	rewind(input);
	m_test_data = new struct svm_node*[total];
	m_target_label = new double[total];
	m_qid = new int[total];
	m_decision_value = new double[total];
	m_test_data_size = total;
	total = 0;

	while(readline(input) != NULL)
	{
		int i = 0;
		char *idx, *val, *label, *endptr;
		int inst_max_index = -1; // strtol gives 0 if wrong format, and precomputed kernel has <index> start from 0
		m_qid[total] = 0;

		label = strtok(line," \t\n");
		if(label == NULL) // empty line
			_exit_input_error(total+1);

		m_target_label[total] = strtod(label,&endptr);
		if(endptr == label || *endptr != '\0')
			_exit_input_error(total+1);

		while(1)
		{
			if(i>=max_nr_attr-1)	// need one more for index = -1
				max_nr_attr *= 2;

			m_test_data[total] = (struct svm_node *) realloc(m_test_data[total],max_nr_attr*sizeof(struct svm_node));
				

			idx = strtok(NULL,":");
			val = strtok(NULL," \t");

			if(val == NULL)
				break;
			if (!strcmp(idx,"qid"))
			{
				m_qid[total] = (int) strtol(val, &endptr,10);
				if(endptr == val || (*endptr != '\0' && !isspace(*endptr)))
					_exit_input_error(i+1);
			}
			else
			{
				m_test_data[total][i].index = (int) strtol(idx,&endptr,10);
				if(endptr == idx || *endptr != '\0' || m_test_data[total][i].index <= inst_max_index)
					_exit_input_error(total+1);
				else
					inst_max_index = m_test_data[total][i].index;

				m_test_data[total][i].value = strtod(val,&endptr);
				if(endptr == val || (*endptr != '\0' && !isspace(*endptr)))
					_exit_input_error(total+1);

				++i;
			}
		}
		m_test_data[total][i].index = -1;
		total++;

	}

	fclose(input);
}

double Logger::get_decision_value(double *w, int index)
{
	double sum = 0.0;
	for (int i = 0; i < m_problem->l; i++)
	{
		svm_node *a = m_problem->x[i], *b = m_test_data[index];
		double _2norm2 = 0.0;
		while (a->index != -1 && b->index != -1)
		{
			if (a->index == b->index)
			{
				double d = a->value - b->value;
				_2norm2 += d * d;

				a++;
				b++;
			}
			else if (a->index < b->index)
				a++;
			else
				b++;
		}

		sum += w[i] * exp(-m_gamma * _2norm2);
	}
	return sum;
}

void Logger::add_record(double *w, double time, double objvalue)
{

	struct Metric record;

#ifndef LOG_TIME_ONLY
	for (int i = 0; i < m_test_data_size; i++)
		m_decision_value[i] = get_decision_value(w, i);
	eval(m_target_label, m_decision_value, m_qid, m_test_data_size, 
			&record.average_pairwise_accuracy, &record.mean_ndcg, &record.mean_average_precision);

	record.objective_function_value = objvalue;
#endif

	record.time = time;

	m_records.push_back(record);
}

void Logger::save(const char *log_file)
{
	FILE *fptr;
	fptr = fopen(log_file, "w");
	assert(fptr != NULL);

	for (int i = 0; i < (int)m_records.size(); i++)
	{
		fprintf(fptr, "iter %d:", i + 1);
		fprintf(fptr," time:%lf", m_records[i].time);

#ifndef LOG_TIME_ONLY
		fprintf(fptr, " obj:%lf", m_records[i].objective_function_value);
		fprintf(fptr, " apa:%lf", m_records[i].average_pairwise_accuracy);
		fprintf(fptr, " meanndcg:%lf", m_records[i].mean_ndcg);
		fprintf(fptr, " map:%lf", m_records[i].mean_average_precision);
#endif
		fprintf(fptr, "\n");
	}
}

